/* =========================================================================
 * File:        FREQMAP.C
 * Description: This module is responsible for mapping between
 *              frequency and frequency table with there named channels.
 *              Please note that named channels have nothing to do with the
 *              broadcasting names. The channels names are standardized names
 *              for frequences. The names changes depending on geographic
 *              location and hence each geographic location uses its own
 *              frequency map to translate to/from numeric frequency values.
 *              By default the "west-europe" map will be used.
 *              All frequences below are given in KHz as integers
 * Author:      Johan Persson (johan162@gmail.com)
 * SVN:         $Id: freqmap.c 831 2014-02-12 02:50:40Z johan162 $
 *
 * Copyright (C) 2009-2014 Johan Persson
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>
 * =========================================================================
 */

// We want the full POSIX and C99 standard
#define _GNU_SOURCE

// And we need to have support for files over 2GB in size
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <sys/param.h> // Needed to get MIN()/MAX()

#include "tvpvrd.h"
#include "tvconfig.h"
#include "freqmap.h"
#include "xstr.h"
#include "utils.h"
#include "tvplog.h"

/*
 * Structure to hold all frequency channel maps. Each maps consists of
 * (table name, number of records, pointer to actual frequency map)
 * This structure gets the pointer to the actual maps resolved
 * at runtime initialization.
 *
 * Each acual frequency map consists of a list of pairs of
 * (Frequency,name) tuples
 *
 * NOTE: The order of this table MUST correspond to the enums in freqm_t
 */
static struct freqmap frequence_map[] = {
    { "europe-west", 0, NULL},
    { "europe-east", 0, NULL},
    { "france", 0, NULL},
    { "ireland", 0, NULL},
    { "italy", 0, NULL},
    { "australia", 0, NULL},
    { "newzealand", 0, NULL},
    { "us-bcast", 0, NULL},
    { "us-cable", 0, NULL},
    { "us-cable-hrc", 0, NULL},
    { "us-cable-irc", 0, NULL},
    { "japan-bcast", 0, NULL},
    { "japan-cable", 0, NULL},
    { "china-bcast", 0, NULL},
    { "southafrica", 0, NULL},
    { "argentina", 0, NULL},
    { "australia-optus", 0, NULL},
    { "" , 0, NULL} /* sentinel */
};

/*
 * Determine size of table of all frequency maps
 */
static const unsigned int num_fmap = sizeof (frequence_map) / sizeof (struct freqmap);

/*
 * Keep a global variable of the map currently in use
 */
static freqm_t curr_fmap=FREQMAP_EUROPEWEST;

/*
 * West Europe standard frequencies.
 * Channel width 7MHz or 8MHz
 */
static struct freqch europe_west_chtable[] = {
    {48250, "E2"},
    {55250, "E3"},
    {62250, "E4"},
    {69250, "S01"},
    {76250, "S02"},
    {83250, "S03"},
    {175250, "E5"},
    {182250, "E6"},
    {189250, "E7"},
    {196250, "E8"},
    {203250, "E9"},
    {210250, "E10"},
    {217250, "E11"},
    {224250, "E12"},
    {105250, "SE1"},
    {112250, "SE2"},
    {119250, "SE3"},
    {126250, "SE4"},
    {133250, "SE5"},
    {140250, "SE6"},
    {147250, "SE7"},
    {154250, "SE8"},
    {161250, "SE9"},
    {168250, "SE10"},
    {231250, "SE11"},
    {238250, "SE12"},
    {245250, "SE13"},
    {252250, "SE14"},
    {259250, "SE15"},
    {266250, "SE16"},
    {273250, "SE17"},
    {280250, "SE18"},
    {287250, "SE19"},
    {294250, "SE20"},
    {303250, "S21"},
    {311250, "S22"},
    {319250, "S23"},
    {327250, "S24"},
    {335250, "S25"},
    {343250, "S26"},
    {351250, "S27"},
    {359250, "S28"},
    {367250, "S29"},
    {375250, "S30"},
    {383250, "S31"},
    {391250, "S32"},
    {399250, "S33"},
    {407250, "S34"},
    {415250, "S35"},
    {423250, "S36"},
    {431250, "S37"},
    {439250, "S38"},
    {447250, "S39"},
    {455250, "S40"},
    {463250, "S41"},
    {471250, "21"},
    {479250, "22"},
    {487250, "23"},
    {495250, "24"},
    {503250, "25"},
    {511250, "26"},
    {519250, "27"},
    {527250, "28"},
    {535250, "29"},
    {543250, "30"},
    {551250, "31"},
    {559250, "32"},
    {567250, "33"},
    {575250, "34"},
    {583250, "35"},
    {591250, "36"},
    {599250, "37"},
    {607250, "38"},
    {615250, "39"},
    {623250, "40"},
    {631250, "41"},
    {639250, "42"},
    {647250, "43"},
    {655250, "44"},
    {663250, "45"},
    {671250, "46"},
    {679250, "47"},
    {687250, "48"},
    {695250, "49"},
    {703250, "50"},
    {711250, "51"},
    {719250, "52"},
    {727250, "53"},
    {735250, "54"},
    {743250, "55"},
    {751250, "56"},
    {759250, "57"},
    {767250, "58"},
    {775250, "59"},
    {783250, "60"},
    {791250, "61"},
    {799250, "62"},
    {807250, "63"},
    {815250, "64"},
    {823250, "65"},
    {831250, "66"},
    {839250, "67"},
    {847250, "68"},
    {855250, "69"}
};

/*
 * French frequency table
 */
static struct freqch france_chtable[] = {
    {47750, "K01"},
    {55750, "K02"},
    {60500, "K03"},
    {63750, "K04"},
    {176000, "K05"},
    {184000, "K06"},
    {192000, "K07"},
    {200000, "K08"},
    {208000, "K09"},
    {216000, "K10"},
    {116750, "KB"},
    {128750, "KC"},
    {140750, "KD"},
    {159750, "KE"},
    {164750, "KF"},
    {176750, "KG"},
    {188750, "KH"},
    {200750, "KI"},
    {212750, "KJ"},
    {224750, "KK"},
    {236750, "KL"},
    {248750, "KM"},
    {260750, "KN"},
    {272750, "KO"},
    {284750, "KP"},
    {296750, "KQ"},
    {303250, "H01"},
    {311250, "H02"},
    {319250, "H03"},
    {327250, "H04"},
    {335250, "H05"},
    {343250, "H06"},
    {351250, "H07"},
    {359250, "H08"},
    {367250, "H09"},
    {375250, "H10"},
    {383250, "H11"},
    {391250, "H12"},
    {399250, "H13"},
    {407250, "H14"},
    {415250, "H15"},
    {423250, "H16"},
    {431250, "H17"},
    {439250, "H18"},
    {447250, "H19"},
    {471250, "21"},
    {479250, "22"},
    {487250, "23"},
    {495250, "24"},
    {503250, "25"},
    {511250, "26"},
    {519250, "27"},
    {527250, "28"},
    {535250, "29"},
    {543250, "30"},
    {551250, "31"},
    {559250, "32"},
    {567250, "33"},
    {575250, "34"},
    {583250, "35"},
    {591250, "36"},
    {599250, "37"},
    {607250, "38"},
    {615250, "39"},
    {623250, "40"},
    {631250, "41"},
    {639250, "42"},
    {647250, "43"},
    {655250, "44"},
    {663250, "45"},
    {671250, "46"},
    {679250, "47"},
    {687250, "48"},
    {695250, "49"},
    {703250, "50"},
    {711250, "51"},
    {719250, "52"},
    {727250, "53"},
    {735250, "54"},
    {743250, "55"},
    {751250, "56"},
    {759250, "57"},
    {767250, "58"},
    {775250, "59"},
    {783250, "60"},
    {791250, "61"},
    {799250, "62"},
    {807250, "63"},
    {815250, "64"},
    {823250, "65"},
    {831250, "66"},
    {839250, "67"},
    {847250, "68"},
    {855250, "69"}
};

/*
 * East european frequency table
 */
static struct freqch europe_east_chtable[] = {
    {49750,"R1"},
    {59250,"R2"},
    {77250,"R3"},
    {85250,"R4"},
    {93250,"R5"},
    {175250,"R6"},
    {183250,"R7"},
    {191250,"R8"},
    {199250,"R9"},
    {207250,"R10"},
    {215250,"R11"},
    {223250,"R12"},
    {111250,"SR1"},
    {119250,"SR2"},
    {127250,"SR3"},
    {135250,"SR4"},
    {143250,"SR5"},
    {151250,"SR6"},
    {159250,"SR7"},
    {167250,"SR8"},
    {231250,"SR11"},
    {239250,"SR12"},
    {247250,"SR13"},
    {255250,"SR14"},
    {263250,"SR15"},
    {271250,"SR16"},
    {279250,"SR17"},
    {287250,"SR18"},
    {295250,"SR19"},
    {48250,"E2"},
    {55250,"E3"},
    {62250,"E4"},
    {69250,"S01"},
    {76250,"S02"},
    {83250,"S03"},
    {175250,"E5"},
    {182250,"E6"},
    {189250,"E7"},
    {196250,"E8"},
    {203250,"E9"},
    {210250,"E10"},
    {217250,"E11"},
    {224250,"E12"},
    {105250,"SE1"},
    {112250,"SE2"},
    {119250,"SE3"},
    {126250,"SE4"},
    {133250,"SE5"},
    {140250,"SE6"},
    {147250,"SE7"},
    {154250,"SE8"},
    {161250,"SE9"},
    {168250,"SE10"},
    {231250,"SE11"},
    {238250,"SE12"},
    {245250,"SE13"},
    {252250,"SE14"},
    {259250,"SE15"},
    {266250,"SE16"},
    {273250,"SE17"},
    {280250,"SE18"},
    {287250,"SE19"},
    {294250,"SE20"},
    {303250,"S21"},
    {311250,"S22"},
    {319250,"S23"},
    {327250,"S24"},
    {335250,"S25"},
    {343250,"S26"},
    {351250,"S27"},
    {359250,"S28"},
    {367250,"S29"},
    {375250,"S30"},
    {383250,"S31"},
    {391250,"S32"},
    {399250,"S33"},
    {407250,"S34"},
    {415250,"S35"},
    {423250,"S36"},
    {431250,"S37"},
    {439250,"S38"},
    {447250,"S39"},
    {455250,"S40"},
    {463250,"S41"},
    {471250,"21"},
    {479250,"22"},
    {487250,"23"},
    {495250,"24"},
    {503250,"25"},
    {511250,"26"},
    {519250,"27"},
    {527250,"28"},
    {535250,"29"},
    {543250,"30"},
    {551250,"31"},
    {559250,"32"},
    {567250,"33"},
    {575250,"34"},
    {583250,"35"},
    {591250,"36"},
    {599250,"37"},
    {607250,"38"},
    {615250,"39"},
    {623250,"40"},
    {631250,"41"},
    {639250,"42"},
    {647250,"43"},
    {655250,"44"},
    {663250,"45"},
    {671250,"46"},
    {679250,"47"},
    {687250,"48"},
    {695250,"49"},
    {703250,"50"},
    {711250,"51"},
    {719250,"52"},
    {727250,"53"},
    {735250,"54"},
    {743250,"55"},
    {751250,"56"},
    {759250,"57"},
    {767250,"58"},
    {775250,"59"},
    {783250,"60"},
    {791250,"61"},
    {799250,"62"},
    {807250,"63"},
    {815250,"64"},
    {823250,"65"},
    {831250,"66"},
    {839250,"67"},
    {847250,"68"},
    {855250,"69"}
};

/*
 * Irish frequency table
 */
static struct freqch ireland_chtable[] = {
    {45750, "A0"},
    {48000, "A1"},
    {53750, "A2"},
    {56000, "A3"},
    {61750, "A4"},
    {64000, "A5"},
    {175250, "A6"},
    {176000, "A7"},
    {183250, "A8"},
    {184000, "A9"},
    {191250, "A10"},
    {192000, "A11"},
    {199250, "A12"},
    {200000, "A13"},
    {207250, "A14"},
    {208000, "A15"},
    {215250, "A16"},
    {216000, "A17"},
    {224000, "A18"},
    {232000, "A19"},
    {248000, "A20"},
    {256000, "A21"},
    {264000, "A22"},
    {272000, "A23"},
    {280000, "A24"},
    {288000, "A25"},
    {296000, "A26"},
    {304000, "A27"},
    {312000, "A28"},
    {320000, "A29"},
    {344000, "A30"},
    {352000, "A31"},
    {408000, "A32"},
    {416000, "A33"},
    {448000, "A34"},
    {480000, "A35"},
    {520000, "A36"},
    {471250, "21"},
    {479250, "22"},
    {487250, "23"},
    {495250, "24"},
    {503250, "25"},
    {511250, "26"},
    {519250, "27"},
    {527250, "28"},
    {535250, "29"},
    {543250, "30"},
    {551250, "31"},
    {559250, "32"},
    {567250, "33"},
    {575250, "34"},
    {583250, "35"},
    {591250, "36"},
    {599250, "37"},
    {607250, "38"},
    {615250, "39"},
    {623250, "40"},
    {631250, "41"},
    {639250, "42"},
    {647250, "43"},
    {655250, "44"},
    {663250, "45"},
    {671250, "46"},
    {679250, "47"},
    {687250, "48"},
    {695250, "49"},
    {703250, "50"},
    {711250, "51"},
    {719250, "52"},
    {727250, "53"},
    {735250, "54"},
    {743250, "55"},
    {751250, "56"},
    {759250, "57"},
    {767250, "58"},
    {775250, "59"},
    {783250, "60"},
    {791250, "61"},
    {799250, "62"},
    {807250, "63"},
    {815250, "64"},
    {823250, "65"},
    {831250, "66"},
    {839250, "67"},
    {847250, "68"},
    {855250, "69"}
};

/*
 * Italian frequency table
 */
static struct freqch italy_chtable[] = {
    {53750, "A"},
    {62250, "B"},
    {82250, "C"},
    {175250, "D"},
    {183750, "E"},
    {192250, "F"},
    {201250, "G"},
    {210250, "H"},
    {217250, "H1"},
    {224250, "H2"},
    {471250, "21"},
    {479250, "22"},
    {487250, "23"},
    {495250, "24"},
    {503250, "25"},
    {511250, "26"},
    {519250, "27"},
    {527250, "28"},
    {535250, "29"},
    {543250, "30"},
    {551250, "31"},
    {559250, "32"},
    {567250, "33"},
    {575250, "34"},
    {583250, "35"},
    {591250, "36"},
    {599250, "37"},
    {607250, "38"},
    {615250, "39"},
    {623250, "40"},
    {631250, "41"},
    {639250, "42"},
    {647250, "43"},
    {655250, "44"},
    {663250, "45"},
    {671250, "46"},
    {679250, "47"},
    {687250, "48"},
    {695250, "49"},
    {703250, "50"},
    {711250, "51"},
    {719250, "52"},
    {727250, "53"},
    {735250, "54"},
    {743250, "55"},
    {751250, "56"},
    {759250, "57"},
    {767250, "58"},
    {775250, "59"},
    {783250, "60"},
    {791250, "61"},
    {799250, "62"},
    {807250, "63"},
    {815250, "64"},
    {823250, "65"},
    {831250, "66"},
    {839250, "67"},
    {847250, "68"},
    {855250, "69"},

};

/*
 * Australian frequency table
 */
static struct freqch australia_chtable[] = {
    {46250, "0"},
    {57250, "1"},
    {64250, "2"},
    {86250, "3"},
    {95250, "4"},
    {102250, "5"},
    {138250, "5A"},
    {175250, "6"},
    {182250, "7"},
    {189250, "8"},
    {196250, "9"},
    {209250, "10"},
    {216250, "11"},
    {527250, "28"},
    {534250, "29"},
    {541250, "30"},
    {548250, "31"},
    {555250, "32"},
    {562250, "33"},
    {569250, "34"},
    {576250, "35"},
    {591250, "36"},
    {604250, "39"},
    {611250, "40"},
    {618250, "41"},
    {625250, "42"},
    {632250, "43"},
    {639250, "44"},
    {646250, "45"},
    {653250, "46"},
    {660250, "47"},
    {667250, "48"},
    {674250, "49"},
    {681250, "50"},
    {688250, "51"},
    {695250, "52"},
    {702250, "53"},
    {709250, "54"},
    {716250, "55"},
    {723250, "56"},
    {730250, "57"},
    {737250, "58"},
    {744250, "59"},
    {751250, "60"},
    {758250, "61"},
    {765250, "62"},
    {772250, "63"},
    {779250, "64"},
    {786250, "65"},
    {793250, "66"},
    {800250, "67"},
    {807250, "68"},
    {814250, "69"},

};

/*
 * New zealand frequency table
 */
static struct freqch newzealand_chtable[] = {
    {45250, "1"},
    {55250, "2"},
    {62250, "3"},
    {175250, "4"},
    {182250, "5"},
    {189250, "6"},
    {196250, "7"},
    {203250, "8"},
    {210250, "9"},
    {217250, "10"},
    {224250, "11"},
    {471250, "21"},
    {479250, "22"},
    {487250, "23"},
    {495250, "24"},
    {503250, "25"},
    {511250, "26"},
    {519250, "27"},
    {527250, "28"},
    {535250, "29"},
    {543250, "30"},
    {551250, "31"},
    {559250, "32"},
    {567250, "33"},
    {575250, "34"},
    {583250, "35"},
    {591250, "36"},
    {599250, "37"},
    {607250, "38"},
    {615250, "39"},
    {623250, "40"},
    {631250, "41"},
    {639250, "42"},
    {647250, "43"},
    {655250, "44"},
    {663250, "45"},
    {671250, "46"},
    {679250, "47"},
    {687250, "48"},
    {695250, "49"},
    {703250, "50"},
    {711250, "51"},
    {719250, "52"},
    {727250, "53"},
    {735250, "54"},
    {743250, "55"},
    {751250, "56"},
    {759250, "57"},
    {767250, "58"},
    {775250, "59"},
    {783250, "60"},
    {791250, "61"},
    {799250, "62"},
    {807250, "63"},
    {815250, "64"},
    {823250, "65"},
    {831250, "66"},
    {839250, "67"},
    {847250, "68"},
    {855250, "69"},
};

/*
 * US Broadcast frequency table
 */
static struct freqch usbcast_chtable[] = {
    {55250, "2"},
    {61250, "3"},
    {67250, "4"},
    {77250, "5"},
    {83250, "6"},
    {175250, "7"},
    {181250, "8"},
    {187250, "9"},
    {193250, "10"},
    {199250, "11"},
    {205250, "12"},
    {211250, "13"},
    {471250, "14"},
    {477250, "15"},
    {483250, "16"},
    {489250, "17"},
    {495250, "18"},
    {501250, "19"},
    {507250, "20"},
    {513250, "21"},
    {519250, "22"},
    {525250, "23"},
    {531250, "24"},
    {537250, "25"},
    {543250, "26"},
    {549250, "27"},
    {555250, "28"},
    {561250, "29"},
    {567250, "30"},
    {573250, "31"},
    {579250, "32"},
    {585250, "33"},
    {591250, "34"},
    {597250, "35"},
    {603250, "36"},
    {609250, "37"},
    {615250, "38"},
    {621250, "39"},
    {627250, "40"},
    {633250, "41"},
    {639250, "42"},
    {645250, "43"},
    {651250, "44"},
    {657250, "45"},
    {663250, "46"},
    {669250, "47"},
    {675250, "48"},
    {681250, "49"},
    {687250, "50"},
    {693250, "51"},
    {699250, "52"},
    {705250, "53"},
    {711250, "54"},
    {717250, "55"},
    {723250, "56"},
    {729250, "57"},
    {735250, "58"},
    {741250, "59"},
    {747250, "60"},
    {753250, "61"},
    {759250, "62"},
    {765250, "63"},
    {771250, "64"},
    {777250, "65"},
    {783250, "66"},
    {789250, "67"},
    {795250, "68"},
    {801250, "69"},
    {807250, "70"},
    {813250, "71"},
    {819250, "72"},
    {825250, "73"},
    {831250, "74"},
    {837250, "75"},
    {843250, "76"},
    {849250, "77"},
    {855250, "78"},
    {861250, "79"},
    {867250, "80"},
    {873250, "81"},
    {879250, "82"},
    {885250, "83"},

};

/*
 * US Cable frequency table
 */
static struct freqch uscable_chtable[] = {
    {73250, "1"},
    {55250, "2"},
    {61250, "3"},
    {67250, "4"},
    {77250, "5"},
    {83250, "6"},
    {175250, "7"},
    {181250, "8"},
    {187250, "9"},
    {193250, "10"},
    {199250, "11"},
    {205250, "12"},
    {211250, "13"},
    {121250, "14"},
    {127250, "15"},
    {133250, "16"},
    {139250, "17"},
    {145250, "18"},
    {151250, "19"},
    {157250, "20"},
    {163250, "21"},
    {169250, "22"},
    {217250, "23"},
    {223250, "24"},
    {229250, "25"},
    {235250, "26"},
    {241250, "27"},
    {247250, "28"},
    {253250, "29"},
    {259250, "30"},
    {265250, "31"},
    {271250, "32"},
    {277250, "33"},
    {283250, "34"},
    {289250, "35"},
    {295250, "36"},
    {301250, "37"},
    {307250, "38"},
    {313250, "39"},
    {319250, "40"},
    {325250, "41"},
    {331250, "42"},
    {337250, "43"},
    {343250, "44"},
    {349250, "45"},
    {355250, "46"},
    {361250, "47"},
    {367250, "48"},
    {373250, "49"},
    {379250, "50"},
    {385250, "51"},
    {391250, "52"},
    {397250, "53"},
    {403250, "54"},
    {409250, "55"},
    {415250, "56"},
    {421250, "57"},
    {427250, "58"},
    {433250, "59"},
    {439250, "60"},
    {445250, "61"},
    {451250, "62"},
    {457250, "63"},
    {463250, "64"},
    {469250, "65"},
    {475250, "66"},
    {481250, "67"},
    {487250, "68"},
    {493250, "69"},
    {499250, "70"},
    {505250, "71"},
    {511250, "72"},
    {517250, "73"},
    {523250, "74"},
    {529250, "75"},
    {535250, "76"},
    {541250, "77"},
    {547250, "78"},
    {553250, "79"},
    {559250, "80"},
    {565250, "81"},
    {571250, "82"},
    {577250, "83"},
    {583250, "84"},
    {589250, "85"},
    {595250, "86"},
    {601250, "87"},
    {607250, "88"},
    {613250, "89"},
    {619250, "90"},
    {625250, "91"},
    {631250, "92"},
    {637250, "93"},
    {643250, "94"},
    {91250, "95"},
    {97250, "96"},
    {103250, "97"},
    {109250, "98"},
    {115250, "99"},
    {649250, "100"},
    {655250, "101"},
    {661250, "102"},
    {667250, "103"},
    {673250, "104"},
    {679250, "105"},
    {685250, "106"},
    {691250, "107"},
    {697250, "108"},
    {703250, "109"},
    {709250, "110"},
    {715250, "111"},
    {721250, "112"},
    {727250, "113"},
    {733250, "114"},
    {739250, "115"},
    {745250, "116"},
    {751250, "117"},
    {757250, "118"},
    {763250, "119"},
    {769250, "120"},
    {775250, "121"},
    {781250, "122"},
    {787250, "123"},
    {793250, "124"},
    {799250, "125"},
    {8250, "T7"},
    {14250, "T8"},
    {20250, "T9"},
    {26250, "T10"},
    {32250, "T11"},
    {38250, "T12"},
    {44250, "T13"},
    {50250, "T14"},

};

/*
 * Japan broadcast
 */
static struct freqch japan_bcast_chtable[] = {
    {9125, "1"},
    {9725, "2"},
    {10325, "3"},
    {17125, "4"},
    {17725, "5"},
    {18325, "6"},
    {18925, "7"},
    {19325, "8"},
    {19925, "9"},
    {20525, "10"},
    {21125, "11"},
    {21725, "12"},
    {47125, "13"},
    {47725, "14"},
    {48325, "15"},
    {48925, "16"},
    {49525, "17"},
    {50125, "18"},
    {50725, "19"},
    {51325, "20"},
    {51925, "21"},
    {52525, "22"},
    {53125, "23"},
    {53725, "24"},
    {54325, "25"},
    {54925, "26"},
    {55525, "27"},
    {56125, "28"},
    {56725, "29"},
    {57325, "30"},
    {57925, "31"},
    {58525, "32"},
    {59125, "33"},
    {59725, "34"},
    {60325, "35"},
    {60925, "36"},
    {61525, "37"},
    {62125, "38"},
    {62725, "39"},
    {63325, "40"},
    {63925, "41"},
    {64525, "42"},
    {65125, "43"},
    {65725, "44"},
    {66325, "45"},
    {66925, "46"},
    {67525, "47"},
    {68125, "48"},
    {68725, "49"},
    {69325, "50"},
    {69925, "51"},
    {70525, "52"},
    {71125, "53"},
    {71725, "54"},
    {72325, "55"},
    {72925, "56"},
    {73525, "57"},
    {74125, "58"},
    {74725, "59"},
    {75325, "60"},
    {75925, "61"},
    {76525, "62"}
};

/*
 * Japan cable
 */
static struct freqch japan_cable_chtable[] = {
{10925,"13"},
{11525,"14"},
{12125,"15"},
{12725,"16"},
{13325,"17"},
{13925,"18"},
{14525,"19"},
{15125,"20"},
{15725,"21"},
{16525,"22"},
{22325,"23"},
{23125,"24"},
{23725,"25"},
{24325,"26"},
{24925,"27"},
{25325,"28"},
{25925,"29"},
{26525,"30"},
{27125,"31"},
{27725,"32"},
{28325,"33"},
{28925,"34"},
{29525,"35"},
{30125,"36"},
{30725,"37"},
{31325,"38"},
{31925,"39"},
{32525,"40"},
{33125,"41"},
{33725,"42"},
{34325,"43"},
{34925,"44"},
{35525,"45"},
{36125,"46"},
{36725,"47"},
{37325,"48"},
{37925,"49"},
{38525,"50"},
{39125,"51"},
{39725,"52"},
{40325,"53"},
{40925,"54"},
{41525,"55"},
{42125,"56"},
{42725,"57"},
{43325,"58"},
{43925,"59"},
{44525,"60"},
{45125,"61"},
{45725,"62"},
{46325,"63"}
};

/*
 * Australia Optus
 */
static struct freqch australia_optus_chtable[] = {
    {13825, "1"},
    {14725, "2"},
    {15425, "3"},
    {16125, "4"},
    {16825, "5"},
    {17525, "6"},
    {18225, "7"},
    {18925, "8"},
    {19625, "9"},
    {20925, "10"},
    {21625, "11"},
    {22425, "12"},
    {23125, "13"},
    {23825, "14"},
    {24525, "15"},
    {25225, "16"},
    {25925, "17"},
    {26625, "18"},
    {27325, "19"},
    {28025, "20"},
    {28725, "21"},
    {29425, "22"},
    {30325, "23"},
    {31025, "24"},
    {31725, "25"},
    {32425, "26"},
    {33825, "27"},
    {34525, "28"},
    {35225, "29"},
    {35925, "30"},
    {36625, "31"},
    {37325, "32"},
    {38025, "33"},
    {38725, "34"},
    {39425, "35"},
    {40125, "36"},
    {40825, "37"},
    {41525, "38"},
    {42225, "39"},
    {42925, "40"},
    {43625, "41"},
    {44325, "42"},
    {45025, "43"},
    {45725, "44"},
    {46425, "45"},
    {47125, "46"},
    {47825, "47"},
    {48525, "48"},
    {49225, "49"},
    {49925, "50"},
    {50625, "51"},
    {51325, "52"},
    {52025, "53"},
    {52725, "54"},
    {53425, "55"}

};

/*
 * China broadcast
 */
static struct freqch china_bcast_chtable[] = {
    {4975, "1"},
    {5775, "2"},
    {6575, "3"},
    {7725, "4"},
    {8525, "5"},
    {11225, "6"},
    {12025, "7"},
    {12825, "8"},
    {13625, "9"},
    {14425, "10"},
    {15225, "11"},
    {16025, "12"},
    {16825, "13"},
    {17625, "14"},
    {18425, "15"},
    {19225, "16"},
    {20025, "17"},
    {20825, "18"},
    {21625, "19"},
    {22425, "20"},
    {23225, "21"},
    {24025, "22"},
    {24825, "23"},
    {25625, "24"},
    {26425, "25"},
    {27225, "26"},
    {28025, "27"},
    {28825, "28"},
    {29625, "29"},
    {30425, "30"},
    {31225, "31"},
    {32025, "32"},
    {32825, "33"},
    {33625, "34"},
    {34425, "35"},
    {35225, "36"},
    {36025, "37"},
    {36825, "38"},
    {37625, "39"},
    {38425, "40"},
    {39225, "41"},
    {40025, "42"},
    {40825, "43"},
    {41625, "44"},
    {42425, "45"},
    {43225, "46"},
    {44025, "47"},
    {44825, "48"},
    {45625, "49"},
    {46325, "50"},
    {47125, "51"},
    {47925, "52"},
    {48725, "53"},
    {49525, "54"},
    {50325, "55"},
    {51125, "56"},
    {51925, "57"},
    {52725, "58"},
    {53525, "59"},
    {54325, "60"},
    {55125, "61"},
    {55925, "62"},
    {60725, "63"},
    {61525, "64"},
    {62325, "65"},
    {63125, "66"},
    {63925, "67"},
    {64725, "68"},
    {65525, "69"},
    {66325, "70"},
    {67125, "71"},
    {67925, "72"},
    {68725, "73"},
    {69525, "74"},
    {70325, "75"},
    {71125, "76"},
    {71925, "77"},
    {72725, "78"},
    {73525, "79"},
    {74325, "80"},
    {75125, "81"},
    {75925, "82"},
    {76725, "83"},
    {77525, "84"},
    {78325, "85"},
    {79125, "86"},
    {79925, "87"},
    {80725, "88"},
    {81525, "89"},
    {82325, "90"},
    {83125, "91"},
    {83925, "92"},
    {84725, "93"},
    {85525, "94"}

};


/*
 * South Africa
 */
static struct freqch southafrica_chtable[] = {
    {17525, "4"},
    {18325, "5"},
    {19125, "6"},
    {19925, "7"},
    {20725, "8"},
    {21525, "9"},
    {22325, "10"},
    {23125, "11"},
    {23925, "12"},
    {24725, "13"},
    {47125, "21"},
    {47925, "22"},
    {48725, "23"},
    {49525, "24"},
    {50325, "25"},
    {51125, "26"},
    {51925, "27"},
    {52725, "28"},
    {53525, "29"},
    {54325, "30"},
    {55125, "31"},
    {55925, "32"},
    {56725, "33"},
    {57525, "34"},
    {58325, "35"},
    {59125, "36"},
    {59925, "37"},
    {60725, "38"},
    {61525, "39"},
    {62325, "40"},
    {63125, "41"},
    {63925, "42"},
    {64725, "43"},
    {65525, "44"},
    {66325, "45"},
    {67125, "46"},
    {67925, "47"},
    {68725, "48"},
    {69525, "49"},
    {70325, "50"},
    {71125, "51"},
    {71925, "52"},
    {72725, "53"},
    {73525, "54"},
    {74325, "55"},
    {75125, "56"},
    {75925, "57"},
    {76725, "58"},
    {77525, "59"},
    {78325, "60"},
    {79125, "61"},
    {79925, "62"},
    {80725, "63"},
    {81525, "64"},
    {82325, "65"},
    {83125, "66"},
    {83925, "67"},
    {84725, "68"},
    {85525, "69"}

};


/*
 * us-cable irc
 */
static struct freqch uscable_irc_chtable[] = {
    {7325, "1"},
    {5525, "2"},
    {6125, "3"},
    {6725, "4"},
    {7925, "5"},
    {8525, "6"},
    {17525, "7"},
    {18125, "8"},
    {18725, "9"},
    {19325, "10"},
    {19925, "11"},
    {20525, "12"},
    {21125, "13"},
    {12115, "14"},
    {12715, "15"},
    {13315, "16"},
    {13915, "17"},
    {14515, "18"},
    {15115, "19"},
    {15715, "20"},
    {16315, "21"},
    {16915, "22"},
    {21725, "23"},
    {22325, "24"},
    {22925, "25"},
    {23525, "26"},
    {24125, "27"},
    {24725, "28"},
    {25325, "29"},
    {25925, "30"},
    {26525, "31"},
    {27125, "32"},
    {27725, "33"},
    {28325, "34"},
    {28925, "35"},
    {29525, "36"},
    {30125, "37"},
    {30725, "38"},
    {31325, "39"},
    {31925, "40"},
    {32525, "41"},
    {33125, "42"},
    {33725, "43"},
    {34325, "44"},
    {34925, "45"},
    {35525, "46"},
    {36125, "47"},
    {36725, "48"},
    {37325, "49"},
    {37925, "50"},
    {38525, "51"},
    {39125, "52"},
    {39725, "53"},
    {40325, "54"},
    {40925, "55"},
    {41525, "56"},
    {42125, "57"},
    {42725, "58"},
    {43325, "59"},
    {43925, "60"},
    {44525, "61"},
    {45125, "62"},
    {45725, "63"},
    {46325, "64"},
    {46925, "65"},
    {47525, "66"},
    {48125, "67"},
    {48725, "68"},
    {49325, "69"},
    {49925, "70"},
    {50525, "71"},
    {51125, "72"},
    {51725, "73"},
    {52325, "74"},
    {52925, "75"},
    {53525, "76"},
    {54125, "77"},
    {54725, "78"},
    {55325, "79"},
    {55925, "80"},
    {56525, "81"},
    {57125, "82"},
    {57725, "83"},
    {58325, "84"},
    {58925, "85"},
    {59525, "86"},
    {60125, "87"},
    {60725, "88"},
    {61325, "89"},
    {61925, "90"},
    {62525, "91"},
    {63125, "92"},
    {63725, "93"},
    {64325, "94"},
    {9125, "95"},
    {9725, "96"},
    {10325, "97"},
    {10925, "98"},
    {11525, "99"},
    {64925, "100"},
    {65525, "101"},
    {66125, "102"},
    {66725, "103"},
    {67325, "104"},
    {67925, "105"},
    {68525, "106"},
    {69125, "107"},
    {69725, "108"},
    {70325, "109"},
    {70925, "110"},
    {71525, "111"},
    {72125, "112"},
    {72725, "113"},
    {73325, "114"},
    {73925, "115"},
    {74525, "116"},
    {75125, "117"},
    {75725, "118"},
    {76325, "119"},
    {76925, "120"},
    {77525, "121"},
    {78125, "122"},
    {78725, "123"},
    {79325, "124"},
    {79925, "125"},
    {825, "T7"},
    {1425, "T8"},
    {2025, "T9"},
    {2625, "T10"},
    {3225, "T11"},
    {3825, "T12"},
    {4425, "T13"},
    {5025, "T14"}
};


/*
 * us-cable hrc
 */
static struct freqch uscable_hrc_chtable[] = {
    {7200, "1"},
    {5400, "2"},
    {6000, "3"},
    {6600, "4"},
    {7800, "5"},
    {8400, "6"},
    {17400, "7"},
    {18000, "8"},
    {18600, "9"},
    {19200, "10"},
    {19800, "11"},
    {20400, "12"},
    {21000, "13"},
    {12000, "14"},
    {12600, "15"},
    {13200, "16"},
    {13800, "17"},
    {14400, "18"},
    {15000, "19"},
    {15600, "20"},
    {16200, "21"},
    {16800, "22"},
    {21600, "23"},
    {22200, "24"},
    {22800, "25"},
    {23400, "26"},
    {24000, "27"},
    {24600, "28"},
    {25200, "29"},
    {25800, "30"},
    {26400, "31"},
    {27000, "32"},
    {27600, "33"},
    {28200, "34"},
    {28800, "35"},
    {29400, "36"},
    {30000, "37"},
    {30600, "38"},
    {31200, "39"},
    {31800, "40"},
    {32400, "41"},
    {33000, "42"},
    {33600, "43"},
    {34200, "44"},
    {34800, "45"},
    {35400, "46"},
    {36000, "47"},
    {36600, "48"},
    {37200, "49"},
    {37800, "50"},
    {38400, "51"},
    {39000, "52"},
    {39600, "53"},
    {40200, "54"},
    {40800, "55"},
    {41400, "56"},
    {42000, "57"},
    {42600, "58"},
    {43200, "59"},
    {43800, "60"},
    {44400, "61"},
    {45000, "62"},
    {45600, "63"},
    {46200, "64"},
    {46800, "65"},
    {47400, "66"},
    {48000, "67"},
    {48600, "68"},
    {49200, "69"},
    {49800, "70"},
    {50400, "71"},
    {51000, "72"},
    {51600, "73"},
    {52200, "74"},
    {52800, "75"},
    {53400, "76"},
    {54000, "77"},
    {54600, "78"},
    {55200, "79"},
    {55800, "80"},
    {56400, "81"},
    {57000, "82"},
    {57600, "83"},
    {58200, "84"},
    {58800, "85"},
    {59400, "86"},
    {60000, "87"},
    {60600, "88"},
    {61200, "89"},
    {61800, "90"},
    {62400, "91"},
    {63000, "92"},
    {63600, "93"},
    {64200, "94"},
    {9000, "95"},
    {9600, "96"},
    {10200, "97"},
    {10800, "98"},
    {11400, "99"},
    {64800, "100"},
    {65400, "101"},
    {66000, "102"},
    {66600, "103"},
    {67200, "104"},
    {67800, "105"},
    {68400, "106"},
    {69000, "107"},
    {69600, "108"},
    {70200, "109"},
    {70800, "110"},
    {71400, "111"},
    {72000, "112"},
    {72600, "113"},
    {73200, "114"},
    {73800, "115"},
    {74400, "116"},
    {75000, "117"},
    {75600, "118"},
    {76200, "119"},
    {76800, "120"},
    {77400, "121"},
    {78000, "122"},
    {78600, "123"},
    {79200, "124"},
    {79800, "125"},
    {700, "T7"},
    {1300, "T8"},
    {1900, "T9"},
    {2500, "T10"},
    {3100, "T11"},
    {3700, "T12"},
    {4300, "T13"},
    {4900, "T14"}
};

/*
 * Argentina
 */
static struct freqch argentina_chtable[] = {
    {5625, "001"},
    {6225, "002"},
    {6825, "003"},
    {7825, "004"},
    {8425, "005"},
    {17625, "006"},
    {18225, "007"},
    {18825, "008"},
    {19425, "009"},
    {20025, "010"},
    {20625, "011"},
    {21225, "012"},
    {12225, "013"},
    {12825, "014"},
    {13425, "015"},
    {14025, "016"},
    {14625, "017"},
    {15225, "018"},
    {15825, "019"},
    {16425, "020"},
    {17025, "021"},
    {21825, "022"},
    {22425, "023"},
    {23025, "024"},
    {23625, "025"},
    {24225, "026"},
    {24825, "027"},
    {25425, "028"},
    {26025, "029"},
    {26625, "030"},
    {27225, "031"},
    {27825, "032"},
    {28425, "033"},
    {29025, "034"},
    {29625, "035"},
    {30225, "036"},
    {30825, "037"},
    {31425, "038"},
    {32025, "039"},
    {32625, "040"},
    {33225, "041"},
    {33825, "042"},
    {34425, "043"},
    {35025, "044"},
    {35625, "045"},
    {36225, "046"},
    {36825, "047"},
    {37425, "048"},
    {38025, "049"},
    {38625, "050"},
    {39225, "051"},
    {39825, "052"},
    {40425, "053"},
    {41025, "054"},
    {41625, "055"},
    {42225, "056"},
    {42825, "057"},
    {43425, "058"},
    {44025, "059"},
    {44625, "060"},
    {45225, "061"},
    {45825, "062"},
    {46425, "063"},
    {47025, "064"},
    {47625, "065"},
    {48225, "066"},
    {48825, "067"},
    {49425, "068"},
    {50025, "069"},
    {50625, "070"},
    {51225, "071"},
    {51825, "072"},
    {52425, "073"},
    {53025, "074"},
    {53625, "075"},
    {54225, "076"},
    {54825, "077"},
    {55425, "078"},
    {56025, "079"},
    {56625, "080"},
    {57225, "081"},
    {57825, "082"},
    {58425, "083"},
    {59025, "084"},
    {59625, "085"},
    {60225, "086"},
    {60825, "087"},
    {61425, "088"},
    {62025, "089"},
    {62625, "090"},
    {63225, "091"},
    {63825, "092"},
    {64425, "093"},

};

/**
 * Initialize the frequency table. Must be called in the beginning of the
 * program.
 */
void
initfreqtable(void) {
    frequence_map[FREQMAP_EUROPEWEST].tbl = europe_west_chtable;
    frequence_map[FREQMAP_EUROPEWEST].size = sizeof (europe_west_chtable) / sizeof (struct freqch);

    frequence_map[FREQMAP_EUROPEEAST].tbl = europe_east_chtable;
    frequence_map[FREQMAP_EUROPEEAST].size = sizeof (europe_east_chtable) / sizeof (struct freqch);

    frequence_map[FREQMAP_FRANCE].tbl = france_chtable;
    frequence_map[FREQMAP_FRANCE].size = sizeof (france_chtable) / sizeof (struct freqch);

    frequence_map[FREQMAP_IRELAND].tbl = ireland_chtable;
    frequence_map[FREQMAP_IRELAND].size = sizeof (ireland_chtable) / sizeof (struct freqch);

    frequence_map[FREQMAP_ITALY].tbl = italy_chtable;
    frequence_map[FREQMAP_ITALY].size = sizeof (italy_chtable) / sizeof (struct freqch);

    frequence_map[FREQMAP_AUSTRALIA].tbl = australia_chtable;
    frequence_map[FREQMAP_AUSTRALIA].size = sizeof (australia_chtable) / sizeof (struct freqch);

    frequence_map[FREQMAP_NEWZEALAND].tbl = newzealand_chtable;
    frequence_map[FREQMAP_NEWZEALAND].size = sizeof (newzealand_chtable) / sizeof (struct freqch);

    frequence_map[FREQMAP_USBCAST].tbl = usbcast_chtable;
    frequence_map[FREQMAP_USBCAST].size = sizeof (usbcast_chtable) / sizeof (struct freqch);

    frequence_map[FREQMAP_USCABLE].tbl = uscable_chtable;
    frequence_map[FREQMAP_USCABLE].size = sizeof (uscable_chtable) / sizeof (struct freqch);
    
    frequence_map[FREQMAP_USCABLE_HRC].tbl = uscable_hrc_chtable;
    frequence_map[FREQMAP_USCABLE_HRC].size = sizeof (uscable_hrc_chtable) / sizeof (struct freqch);    

    frequence_map[FREQMAP_USCABLE_IRC].tbl = uscable_irc_chtable;
    frequence_map[FREQMAP_USCABLE_IRC].size = sizeof (uscable_irc_chtable) / sizeof (struct freqch);    
    
    frequence_map[FREQMAP_JAPAN_BCAST].tbl = japan_bcast_chtable;
    frequence_map[FREQMAP_JAPAN_BCAST].size = sizeof (japan_bcast_chtable) / sizeof (struct freqch);    

    frequence_map[FREQMAP_JAPAN_CABLE].tbl = japan_cable_chtable;
    frequence_map[FREQMAP_JAPAN_CABLE].size = sizeof (japan_cable_chtable) / sizeof (struct freqch);    
 
    frequence_map[FREQMAP_CHINA_BCAST].tbl = china_bcast_chtable;
    frequence_map[FREQMAP_CHINA_BCAST].size = sizeof (china_bcast_chtable) / sizeof (struct freqch);    

    frequence_map[FREQMAP_SOUTHAFRICA].tbl = southafrica_chtable;
    frequence_map[FREQMAP_SOUTHAFRICA].size = sizeof (southafrica_chtable) / sizeof (struct freqch);    
 
    frequence_map[FREQMAP_ARGENTINA].tbl = argentina_chtable;
    frequence_map[FREQMAP_ARGENTINA].size = sizeof (argentina_chtable) / sizeof (struct freqch);    
    
    frequence_map[FREQMAP_AUSTRALIA_OPTUS].tbl = australia_optus_chtable;
    frequence_map[FREQMAP_AUSTRALIA_OPTUS].size = sizeof (australia_optus_chtable) / sizeof (struct freqch);    
    
}

struct station_map_entry {
    char name[32];
    char channel[32];
};

static struct station_map_entry station_map[256] ;
static size_t num_stations = 0 ;

/**
 * Try to read broadcast names mapping to channel names
 * @param name
 * @return -1 on failure, 0 on success
 */
int
read_xawtvfile(const char *name) {
    dictionary *rdict;

    rdict = iniparser_load(name);
    if( rdict == NULL ) {
        logmsg(LOG_ERR,"Could not read channel/station alias file '%s' (%d ; %s). ",name,errno,strerror(errno));
        return -1;
    }
    logmsg(LOG_DEBUG,"Reading channel/station alias file '%s'",name);

    // Loop through all sections to find the corresponding channel
    int nsec = iniparser_getnsec(rdict);
    char buffer[64];
    for(int i=0; i < nsec; i++) {
        char *sname = iniparser_getsecname(rdict, i);
        strncpy(buffer,sname,sizeof(buffer)-1);
        strncat(buffer,":channel",sizeof(buffer)-1-strlen(sname));
        buffer[63] = '\0';
        char *channel_name = iniparser_getstring(rdict,buffer,NULL);
        if( channel_name ) {
            strncpy(station_map[num_stations].name,sname,31);
            station_map[num_stations].name[31] = '\0';
            strncpy(station_map[num_stations].channel,channel_name,31);
            station_map[num_stations].channel[31] = '\0';
            
            // Check that the channel really exists in the frequency map specified.
            // This means that set_current_freqmap() has to be called prior to
            // this function
            unsigned int freq;
            logmsg(LOG_DEBUG,"Checking that channel \"%s\" (for station \"%s\") exists in freq. map",channel_name,sname);
            if( -1 == getfreqfromch(&freq, channel_name) ) {
                logmsg(LOG_ERR,"Channel \"%s\" set for station name \"%s\" in '%s' does not exist in current frequency map.",channel_name,sname,name);
                iniparser_freedict(rdict);
                return -1;                
            }
                                   
            num_stations++;
        }
    }    
    iniparser_freedict(rdict);
    logmsg(LOG_NOTICE,"Successfully read station/channel alias file '%s'. Found %d stations.",name,num_stations);
    return 0;
}

void
list_stations(int fd) {
    for(unsigned i=0; i<num_stations; i++) {
        _writef(fd,"%-5s: %s\n",station_map[i].channel,station_map[i].name);
    }
}

int
_qstrcmp(const void *s1, const void *s2) {
    //return strcmp(*(char *)s1,*(char *)s2);
    return strcmp(* (char * const *) s1, * (char * const *) s2);

}

int
get_stations(const char *stations[],size_t maxlen) {
    unsigned n = MIN(num_stations,maxlen);
    for(unsigned i=0; i < n; i++) {
        stations[i] = station_map[i].name;
    }

    // JP 2013-11-10 Comment out sort. Preserve channel 
    // order in mapping file instead since a user might want
    // to sort according to most watched.
    // qsort(stations,n,sizeof(char *),_qstrcmp);

    return (int)n;
}

/**
 * Translate a station name to a channel name
 * @param station
 * @param chbuffer
 * @param size
 * @return 0 on success, -1 on failure
 */
int
get_chfromstation(const char *station, char *chbuffer, size_t size) {
    for(unsigned i=0; i < num_stations; i++) {
        if( xstricmp(station,station_map[i].name) == 0  ) {
            strncpy(chbuffer, station_map[i].channel, size);
            chbuffer[size-1]='\0';
            return 0;
        }
    }
    return -1;
}

/**
 * Set the current frequency map
 * @param name frequency map name
 * @return -1 on failure, idx >= 0 on sucess
 */
int
set_current_freqmap(const char *name) {
    int idx = getfmapidx(name);
    if( idx == -1 )
        return idx;
    curr_fmap = idx;
    logmsg(LOG_NOTICE,"Frequency map set to '%s'",name);
    return idx;
}

/**
 * Get the current frequency map, both name and index
 * @param name Buffer set to the name of the current map
 * @param size Maximum size in bytes of buffer
 * @return The current map index
 */
int
get_current_freqmap(char *name, size_t size) {
    strncpy(name,frequence_map[curr_fmap].name,size);
    name[size-1] = '\0';
    return curr_fmap;
}

/**
 * Get the corresponding frequency map index from a map name
 */
int
getfmapidx(const char *name) {
    for (unsigned i = 0; i < num_fmap; i++) {
        if (xstricmp(name, frequence_map[i].name) == 0  && frequence_map[i].size > 0)
            return (int)i;
    }
    return -1;
}

/**
 * Translate a frequency to the standard channel name in the frequency map.
 * Note: it isthe callers responsibility that the parameter fmap is a valid
 * map.
 * @return 0 on successs, -1 otherwise
 */
int
getchfromfreq(char **ch, const unsigned int freq) {
    if (curr_fmap >= num_fmap) {
        return -1;
    }
    for (unsigned i = 0; i < frequence_map[curr_fmap].size; ++i) {
        if (freq == frequence_map[curr_fmap].tbl[i].freq*1000) {
            *ch = &frequence_map[curr_fmap].tbl[i].ch[0];
            return 0;
        }
    }
    return -1;
}

/**
 * Translate a channel name to a frequency using the specified frequency map.
 * For the special case where the channel name starts with a '#' character interpret
 * the channel name as a frequency directly
 * @return 0 on success, -1 otherwise
 */
int
getfreqfromch(unsigned int *freq, const char *ch) {
    /* Start with sanity check of the current frequency map */
    if (curr_fmap >= num_fmap) {
        return -1;
    }
    
    /* Check if the frequency is specified directly */
    if( '@' == *ch ) {
        
        /* Must be four to six digits to specify frequency */
        size_t n = strnlen(ch,7);
        if( n < 5 || n > 7 ) {
            return -1;
        }
        char nch[7];
        strncpy(nch,ch+1,6);
        *freq = xatoi(nch);
        if( *freq < 1000 || *freq > 999999 ) {
            return -1;
        }
        logmsg(LOG_DEBUG,"Channel frequency manually specified to %u kHz ",*freq);
        *freq *= 1000;
        return 0;
        
    } else {     
        
        for (unsigned int i = 0; i < frequence_map[curr_fmap].size; ++i) {
            if (xstricmp(ch, frequence_map[curr_fmap].tbl[i].ch) == 0 ) {
                *freq = frequence_map[curr_fmap].tbl[i].freq*1000;
                return 0;
            }
        }
        
    }
    return -1;
}

/**
 * Translate a symbolic name to a frequency. The string can be given as
 * either a station or channel name
 * @param freq
 * @param name
 * @return 0 on success, -1 on failure
 */
int
getfreqfromstr(unsigned int *freq,const char *name) {
    char chbuffer[32];
    // First try if this is a station name
    if( 0 == get_chfromstation(name, chbuffer, 32) ) {
        if( 0 == getfreqfromch(freq, chbuffer) ) {
            return 0;
        }
    }
    else {
        if( 0 == getfreqfromch(freq, name) ) {
            return 0;
        }
    }
    return -1;
}

/*
 * Fill the supplied buffer (array of string pointers) 
 * with pointer to statically allocated buffers with the name of 
 * all the defined frequency maps
 */
unsigned
getfmapnames(char *names[], size_t size) {
    unsigned i;
    for(i=0; i < size && i < num_fmap; i++) {
        names[i] = frequence_map[i].name;
    }
    return i;
}

/**
 * Store a list of "maxnum" elements length with pointers to all names of the
 * predefined frequency tables
 * 
 * @param list
 * @param maxnum
 * @return 0 on success, -1 on failure
 */
int
getfreqmaplist(char *list[], size_t maxnum) {
    size_t i=0;
    while( frequence_map[i].size > 0 && i < maxnum ) {
        list[i] = frequence_map[i].name;
        ++i;
    }
    return i > 0 ? 0 : -1;
}

/* freqmap.c */
